'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = convertTypeToPropTypes;

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _babelLog = require('babel-log');

var _babelFileLoader = require('babel-file-loader');

var _matchExported = require('./matchExported');

var _matchExported2 = _interopRequireDefault(_matchExported);

var _error = require('./error');

var _error2 = _interopRequireDefault(_error);

var _babelFlowScope = require('babel-flow-scope');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function cloneComments(comments) {
  return comments && comments.map(function (comment) {
    comment = t.clone(comment);
    comment.start = comment.start + 0.0001; // Force printer to print... (sigh)
    return comment;
  });
}


function inheritsComments(a, b) {
  return t.inheritsComments(a, {
    trailingComments: cloneComments(b.trailingComments),
    leadingComments: cloneComments(b.leadingComments),
    innerComments: cloneComments(b.innerComments)
  });
}

function isLiteralTypeAnnotation(node) {
  return t.isStringLiteralTypeAnnotation(node) || t.isBooleanLiteralTypeAnnotation(node) || t.isNumericLiteralTypeAnnotation(node) || t.isVoidTypeAnnotation(node) || t.isNullLiteralTypeAnnotation(node);
}

function typeToValue(node) {
  var value = void 0;

  if (t.isVoidTypeAnnotation(node)) {
    value = undefined;
  } else if (t.isNullLiteralTypeAnnotation(node)) {
    value = null;
  } else {
    value = node.value;
  }

  return t.valueToNode(value);
}

var refPropTypes = function refPropTypes(property, opts) {
  return t.memberExpression(opts.getPropTypesRef(), property);
};

var createThrows = function createThrows(message) {
  return function (path, opts) {
    throw (0, _error2.default)(path, message);
  };
};

var createConversion = function createConversion(name) {
  return function (path, opts) {
    return refPropTypes(t.identifier(name), opts);
  };
};

var convertLiteral = function convertLiteral(path, opts) {
  var arr = t.arrayExpression([typeToValue(path.node)]);
  return t.callExpression(refPropTypes(t.identifier('oneOf'), opts), [arr]);
};

var converters = {};

converters.AnyTypeAnnotation = createConversion('any');
converters.MixedTypeAnnotation = createConversion('any');
converters.NumberTypeAnnotation = createConversion('number');
converters.NumericLiteralTypeAnnotation = convertLiteral;
converters.BooleanTypeAnnotation = createConversion('bool');
converters.BooleanLiteralTypeAnnotation = convertLiteral;
converters.StringTypeAnnotation = createConversion('string');
converters.StringLiteralTypeAnnotation = convertLiteral;
converters.NullLiteralTypeAnnotation = convertLiteral;
converters.VoidTypeAnnotation = convertLiteral;
converters.FunctionTypeAnnotation = createConversion('func');
converters.TupleTypeAnnotation = createConversion('array');

converters.NullableTypeAnnotation = function (path, opts) {
  return t.callExpression(refPropTypes(t.identifier('oneOf'), opts), [t.valueToNode(null), t.valueToNode(undefined), convert(path.get('typeAnnotation'), opts)]);
};

converters.QualifiedTypeIdentifier = createThrows('qualified type identifiers unsupported');

converters.TypeAnnotation = function (path, opts) {
  return convert(path.get('typeAnnotation'), opts);
};

converters.ObjectTypeAnnotation = function (path, opts) {
  var _path$node = path.node,
      properties = _path$node.properties,
      indexers = _path$node.indexers,
      callProperties = _path$node.callProperties;


  if (properties.length && indexers.length) {
    throw (0, _error2.default)(path, 'Objects with both properties and indexers are unsupported');
  }

  if (!properties.length && indexers.length > 1) {
    throw (0, _error2.default)(path, 'Objects with multiple indexers are unsupported');
  }

  if (callProperties.length) {
    throw (0, _error2.default)(path.get('callProperties')[0], 'Object call properties are unsupported');
  }

  if (indexers.length) {
    var indexer = path.get('indexers')[0];
    return t.callExpression(refPropTypes(t.identifier('objectOf'), opts), [convert(indexer, opts)]);
  } else {
    var props = [];

    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = path.get('properties')[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var property = _step.value;

        props.push(convert(property, opts));
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    var object = t.objectExpression(props);

    return t.callExpression(refPropTypes(t.identifier('shape'), opts), [object]);
  }
};

converters.ObjectTypeProperty = function (path, opts) {
  var key = path.get('key');
  var value = path.get('value');

  var id = void 0;
  if (key.isStringLiteral()) {
    id = t.stringLiteral(key.node.value);
  } else {
    id = t.identifier(key.node.name);
  }

  var converted = convert(value, opts);

  if (!path.node.optional) {
    converted = t.memberExpression(converted, t.identifier('isRequired'));
  }

  return t.objectProperty(inheritsComments(id, key.node), converted);
};

converters.ObjectTypeIndexer = function (path, opts) {
  return convert(path.get('value'), opts);
};

converters.ArrayTypeAnnotation = function (path, opts) {
  return t.callExpression(refPropTypes(t.identifier('arrayOf'), opts), [convert(path.get('elementType'), opts)]);
};

var typeParametersConverters = {
  Array: function Array(path, opts) {
    var param = path.get('typeParameters').get('params')[0];
    return t.callExpression(refPropTypes(t.identifier('arrayOf'), opts), [convert(param, opts)]);
  }
};

converters.GenericTypeAnnotation = function (path, opts) {
  if (!path.node.typeParameters) {
    return convert(path.get('id'), opts);
  }

  var name = path.node.id.name;

  if (typeParametersConverters[name]) {
    return typeParametersConverters[name](path, opts);
  } else {
    throw (0, _error2.default)(path, 'Unsupported generic type annotation with type parameters');
  }
};

var typeIdentifierConverters = {
  Function: createConversion('func'),
  Object: createConversion('object')
};

converters.Identifier = function (path, opts) {
  var name = path.node.name;

  if (path.parentPath.isFlow() && typeIdentifierConverters[name]) {
    return typeIdentifierConverters[name](path, opts);
  }

  var binding = (0, _babelFlowScope.findFlowBinding)(path, name);

  if (!binding) {
    throw (0, _error2.default)(path, 'Missing reference "' + name + '"');
  }

  var bindingPath = void 0;

  if (binding.kind === 'declaration') {
    bindingPath = binding.path.parentPath;
  } else if (binding.kind === 'import') {
    bindingPath = binding.path.parentPath;
  } else if (binding.kind === 'param') {
    throw binding.path.buildCodeFrameError('Cannot convert type parameters');
  } else {
    throw new Error('Unexpected Flow binding kind: ' + binding.kind);
  }

  return convert(bindingPath, opts);
};

converters.TypeAlias = function (path, opts) {
  return convert(path.get('right'), opts);
};

converters.InterfaceDeclaration = function (path, opts) {
  return convert(path.get('body'), opts);
};

converters.ClassDeclaration = function (path, opts, id) {
  return t.callExpression(refPropTypes(t.identifier('instanceOf'), opts), [id || t.identifier(path.node.id.name)]);
};

converters.UnionTypeAnnotation = function (path, opts) {
  var isLiterals = path.node.types.every(isLiteralTypeAnnotation);
  var propType = void 0;
  var elements = void 0;

  if (isLiterals) {
    propType = 'oneOf';
    elements = path.get('types').map(function (p) {
      return typeToValue(p.node);
    });
  } else {
    propType = 'oneOfType';
    elements = path.get('types').map(function (p) {
      return convert(p, opts);
    });
  }

  var arr = t.arrayExpression(elements);
  return t.callExpression(refPropTypes(t.identifier(propType), opts), [arr]);
};

converters.IntersectionTypeAnnotation = function (path, opts) {
  return t.callExpression(opts.getPropTypesAllRef(), path.get('types').map(function (type) {
    return convert(type, opts);
  }));
};

function _convertImportSpecifier(path, opts) {
  var kind = path.parent.importKind;
  if (kind === 'typeof') {
    throw (0, _error2.default)(path, 'import typeof is unsupported');
  }

  var file = (0, _babelFileLoader.loadImportSync)(path.parentPath, opts.resolveOpts);
  var local = path.node.local.name;
  var name = void 0;

  if (path.type === 'ImportDefaultSpecifier' && kind === 'value') {
    name = 'default';
  } else {
    name = local;
  }

  var exported = (0, _matchExported2.default)(file, name);

  if (!exported) {
    throw (0, _error2.default)(path, 'Missing matching export');
  }

  return convert(exported, opts, t.identifier(name));
}

converters.ImportDefaultSpecifier = function (path, opts) {
  return _convertImportSpecifier(path, opts);
};

converters.ImportSpecifier = function (path, opts) {
  return _convertImportSpecifier(path, opts);
};

var convert = function convert(path, opts, id) {
  var converter = converters[path.type];

  if (!converter) {
    throw (0, _error2.default)(path, 'No converter for node type: ' + path.type);
  }

  return inheritsComments(converter(path, opts, id), path.node);
};

function convertTypeToPropTypes(typeAnnotation, opts) {
  return convert(typeAnnotation, opts).arguments[0];
}